"""
This backend is based on OpenDX (www.opendx.org) which is the open source
version of IBM's Visualization Data Explorer and is a powerfull, full-
featured software package for visualization of scientific, engineering
and analytical data. The connection with Python is handled by py2dx which
is available for download at http://www.psc.edu/~eschenbe/. This backend
can be used by

  python somefile.py --SCITOOLS_easyviz_backend dx

or one can specify the backend in the SciTools configuration file
scitools.cfg under the [easyviz] section

  [easyviz]
  backend = dx

and then

  from scitools.std import *

or if just easyviz is needed

  from scitools.easyviz import *

REQUIREMENTS:

OpenDX 
py2dx 

Tips:

* The following is an example on how one can send commands directly to DX:

  g = get_backend()
  g('<your DX command here>')

* The complete DX script can be saved to disk for running at a later stage.
  See save_script and its doc string for further details.

TODO:

* Set axis limits. Use the ClipBox module. Problems with AutoAxes.

* Add support for vector fields in two and three dimensions. In 3D we can
  create the vector field by

  f = open('vector3D.dat', 'w')
  nx, ny = shape(z)
  for i in iseq(nx-1):
      for j in iseq(ny-1):
          f.write('%5.3lf\t%5.3lf\t%5.3lf\t%5.3lf\t%5.3lf\t%5.3lf\n' % \
                  (x[i,j],y[i,j],z[i,j],u[i,j],v[i,j],w[i,j]))
  f.close()

  and then use a general header file like this

    file = vector3D.dat
    points = 25
    format = ascii
    interleaving = field
    field = locations, field0
    structure = 3-vector, 3-vector
    type = float, float
    end

  Then use the AutoGlyph/Glyph module in DX to create arrows or use
  Streamline alone or together with Ribbon/Tube.
  (Similar applies to 2D vector fields)

* Fix problems with camera.

* Look more into scaling of data fields.

* Implement different colormaps.

* Is it possible to implement the coneplot command?

* Add support for the isocaps command:

  isocaps = CappedIsosurfaceMacro(data,isovalue,capDir=0);

  Problem: Not only isocaps but also complete isosurface.

* How can we treat NaNs? 
  (see http://www.opendx.org/cgi-bin/forum/YaBB.pl?num=1155198504)

* Let _DXFigure be based on DXServer rather than just DX.

* In addition look at curve plotting (Plot module), caxis, colorbar,
  meshc/surfc, ...

"""


from __future__ import division

from common import *
from scitools.globaldata import DEBUG, VERBOSE
from scitools.misc import test_if_module_exists 

test_if_module_exists('DX', msg='You need to install the py2dx package.', abort=False)
import DX
import Tkinter
import tempfile


DXMACROS = '/usr/share/dx/samples/macros'
DXMACROS = os.environ.get('DXMACROS', DXMACROS)


class _DXFigure(object):
    def __init__(self, plt, width=640, height=480, depth=24, title='',
                 hw_render_mode='opengl'):
        # create the GUI:
        self.plt = plt
        self.width = width
        self.height = height
        self.master = plt._master
        self.root = Tkinter.Toplevel(self.master)
        self.root.title(title)
        self.root.protocol("WM_DELETE_WINDOW", self.close)
        self.root.bind("<KeyPress-q>", self.close)
        self.root.minsize(200, 200)
        self.root.geometry('%sx%s' % (width,height))
        self.root.withdraw()
        self.frame = Tkinter.Frame(self.root, relief='sunken', bd=2)
        self.frame.pack(side='top', fill='both', expand=1)
        #button = Tkinter.Button(self.root, text="Quit", command=self.close)
        #button.pack()

        # start DX:
        cmd = DX.DXEXEC + ' -execonly -hwrender opengl'
        self.conn = DX.DXLStartDX(cmd, None)

        #DX.DXLLoadMacroDirectory(self.conn, DXMACROS)
        # hack for loading necessary macro files:
        DX.exDXLLoadScript(self.conn,
                           os.path.join(DXMACROS, "ArrangeMemberMacro.net"))
        DX.exDXLLoadScript(self.conn,
                           os.path.join(DXMACROS, "CappedIsoMacro.net"))

        self.script = ''
                
    def __call__(self, cmd):
        self.send(cmd)

    def send(self, cmd):
        if not isinstance(cmd, str):
            raise ValueError('DX command must be string, not %s' \
                             % type(cmd))
        self.script += cmd + '\n'
        if not cmd.startswith('//'):  # don't send comments to DX
            DX.DXLSend(self.conn, cmd)

    def reset(self):
        self.script = ''
        DX.DXLEndExecuteOnChange(self.conn)
        DX.DXLSync(self.conn)

    def close(self, event=None):
        self.plt.clf()
        self.root.withdraw()

    def display(self, show=True):
        if show:
            self.root.deiconify()
        self.root.update()
        DX.DXLExecuteOnChange(self.conn)
        DX.WaitForDXIdle(self.conn)
        #DX.DXLExecuteOnce(self.conn)
        #DX.uiDXLOpenVPE(self.conn)

    def exit(self):
        self.root.destroy()
        DX.WaitForDXIdle(self.conn)
        DX.DXLExitDX(self.conn)


class DXBackend(BaseClass):
    def __init__(self):
        BaseClass.__init__(self)
        self._init()
        
    def _init(self, *args, **kwargs):
        """Perform initialization that is special for this backend."""
        
        self._master = Tkinter.Tk()
        self._master.withdraw()
        self.figure(self.getp('curfig'))

        # conversion tables for format strings:
        self._markers = {
            '': None,   # no marker
            '.': None,  # dot
            'o': None,  # circle
            'x': None,  # cross
            '+': None,  # plus sign
            '*': None,  # asterisk
            's': None,  # square
            'd': None,  # diamond
            '^': None,  # triangle (up)
            'v': None,  # triangle (down)
            '<': None,  # triangle (left)
            '>': None,  # triangle (right)
            'p': None,  # pentagram
            'h': None,  # hexagram
            }
                        
        self._colors = {
            '': None,       # no color --> blue
            'r': "red",     # red
            'g': "green",   # green
            'b': "blue",    # blue
            'c': "cyan",    # cyan
            'm': "magenta", # magenta
            'y': "yellow",  # yellow
            'k': "black",   # black
            'w': "white",   # white
            }
        
        self._line_styles = {
            '': None,    # no line 
            '-': None,   # solid line
            ':': None,   # dotted line
            '-.': None,  # dash-dot line
            '--': None,  # dashed line
            }

        # convert table for colorbar location:
        self._colorbar_locations = {
            'North': None,
            'South': None,
            'East': None,
            'West': None,
            'NorthOutside': None,
            'SouthOutside': None,
            'EastOutside': None,
            'WestOutside': None,
            }

        if DEBUG:
            print "Setting backend standard variables"
            for disp in 'self._markers self._colors self._line_styles'.split():
                print disp, eval(disp)

    def _set_scale(self, ax):
        """Set linear or logarithmic (base 10) axis scale."""
        if DEBUG:
            print "Setting scales"
        scale = ax.getp('scale')
        if scale == 'loglog':
            # use logarithmic scale on both x- and y-axis
            pass
        elif scale == 'logx':
            # use logarithmic scale on x-axis and linear scale on y-axis
            pass
        elif scale == 'logy':
            # use linear scale on x-axis and logarithmic scale on y-axis
            pass
        elif scale == 'linear':
            # use linear scale on both x- and y-axis
            pass

    def _set_labels(self, ax):
        """Add text labels for x-, y-, and z-axis."""
        if DEBUG:
            print "Setting labels"
        xlabel = ax.getp('xlabel')
        ylabel = ax.getp('ylabel')
        zlabel = ax.getp('zlabel')
        self._g('labels = {"%s","%s","%s"};' % (xlabel,ylabel,zlabel))
        
    def _set_title(self, ax):
        """Add a title at the top of the axis."""
        if DEBUG:
            print "Setting title"
        title = self._fix_latex(ax.getp('title'))
        self._g('title = "%s";' % title)
        self._g('caption = Caption(title,[0.5,0.98],font="fixed");')
        self._g('collected = Append(collected,caption);')
    
    def _set_limits(self, ax):
        """Set axis limits in x, y, and z direction."""
        if DEBUG:
            print "Setting axis limits"
        mode = ax.getp('mode')
        if mode == 'auto':
            # let plotting package set 'nice' axis limits in the x, y,
            # and z direction. If this is not automated in the plotting
            # package, one can use the following limits:
            xmin, xmax, ymin, ymax, zmin, zmax = ax.get_limits()
        elif mode == 'manual':
            # (some) axis limits are frozen
            xmin = ax.getp('xmin')
            xmax = ax.getp('xmax')
            if xmin is not None and xmax is not None:
                # set x-axis limits
                pass
            else:
                # let plotting package set x-axis limits or use
                xmin, xmax = ax.getp('xlim')

            ymin = ax.getp('ymin')
            ymax = ax.getp('ymax')
            if ymin is not None and ymax is not None:
                # set y-axis limits
                pass
            else:
                # let plotting package set y-axis limits or use
                ymin, ymax = ax.getp('ylim')

            zmin = ax.getp('zmin')
            zmax = ax.getp('zmax')
            if zmin is not None and zmax is not None:
                # set z-axis limits
                pass
            else:
                # let plotting package set z-axis limits or use
                zmin, zmax = ax.getp('zlim')
        elif mode == 'tight':
            # set the limits on the axis to the range of the data. If
            # this is not automated in the plotting package, one can
            # use the following limits:
            xmin, xmax, ymin, ymax, zmin, zmax = ax.get_limits()
        elif mode == 'fill':
            # not sure about this
            xmin, xmax, ymin, ymax, zmin, zmax = ax.get_limits()

        # scale limits according to data aspect ratio:
        dar = ax.getp('daspect')
        xmin /= dar[0];  xmax /= dar[0]
        ymin /= dar[1];  ymax /= dar[1]
        zmin /= dar[2];  zmax /= dar[2]
        # create vector to be used in AutoAxes later:
        self._g('axislimits = {[%s,%s,%s],[%s,%s,%s]};' % \
                (xmin,ymin,zmin,xmax,ymax,zmax))
        #self._g('clipped = ClipBox(collected,axislimits);')
        #self._g('collected = Collect(clipped);')

    def _set_position(self, ax):
        """Set axes position."""
        rect = ax.getp('viewport')
        if rect:
            # axes position is defined. In Matlab rect is defined as
            # [left,bottom,width,height], where the four parameters are
            # location values between 0 and 1 ((0,0) is the lower-left
            # corner and (1,1) is the upper-right corner).
            # NOTE: This can be different in the plotting package.
            pass

    def _set_daspect(self, ax):
        """Set data aspect ratio."""
        if ax.getp('daspectmode') == 'manual':
            dar = ax.getp('daspect')  # dar is a list (len(dar) is 3).
            self._g('daspect = [%s %s %s];' % (dar[0],dar[1],dar[2]))
        else:
            # daspectmode is 'auto'. Plotting package handles data
            # aspect ratio automatically.
            pass
        
    def _set_axis_method(self, ax):
        method = ax.getp('method')
        if method == 'equal':
            # tick mark increments on the x-, y-, and z-axis should
            # be equal in size.
            pass
        elif method == 'image':
            # same effect as axis('equal') and axis('tight')
            pass
        elif method == 'square':
            # make the axis box square in size
            pass
        elif method == 'normal':
            # full size axis box
            pass
        elif method == 'vis3d':
            # freeze data aspect ratio when rotating 3D objects
            pass

    def _set_coordinate_system(self, ax):
        """
        Use either the default Cartesian coordinate system or a
        matrix coordinate system.
        """
        
        direction = ax.getp('direction')
        if direction == 'ij':
            # Use matrix coordinates. The origin of the coordinate
            # system is the upper-left corner. The i-axis should be
            # vertical and numbered from top to bottom, while the j-axis
            # should be horizontal and numbered from left to right.
            pass
        elif direction == 'xy':
            # use the default Cartesian axes form. The origin is at the
            # lower-left corner. The x-axis is vertical and numbered
            # from left to right, while the y-axis is vertical and
            # numbered from bottom to top.
            pass

    def _set_box(self, ax):
        """Turn box around axes boundary on or off."""
        if DEBUG:
            print "Setting box"
        if ax.getp('box'):
            # display box 
            self._g('show_box = 1;')
        else:
            # do not display box
            self._g('show_box = 0;')
        
    def _set_grid(self, ax):
        """Turn grid lines on or off."""
        if DEBUG:
            print "Setting grid"
        if ax.getp('grid'):
            # turn grid lines on
            self._g('show_grid = 1;')
        else:
            # turn grid lines off
            self._g('show_grid = 0;')

    def _set_hidden_line_removal(self, ax):
        """Turn on/off hidden line removal for meshes."""
        if DEBUG:
            print "Setting hidden line removal"
        if ax.getp('hidden'):
            # turn hidden line removal on
            pass
        else:
            # turn hidden line removal off
            pass

    def _set_colorbar(self, ax):
        """Add a colorbar to the axis."""
        if DEBUG:
            print "Setting colorbar"
        cbar = ax.getp('colorbar')
        if cbar.getp('visible'):
            # turn on colorbar
            cbar_title = cbar.getp('cbtitle')
            cbar_location = self._colorbar_locations[cbar.getp('cblocation')]
            #xpos, ypos, length, width, horiz = cbar_location
            #self._g('cbar = ColorBar(collected);')
            #self._g('collected = Append(collected,cbar);')
        else:
            # turn off colorbar
            pass

    def _set_caxis(self, ax):
        """Set the color axis scale."""
        if DEBUG:
            print "Setting caxis"
        if ax.getp('caxismode') == 'manual':
            cmin, cmax = ax.getp('caxis')
            # NOTE: cmin and cmax might be None:
            if cmin is None or cmax is None:
                cmin, cmax = [0,1]
            # set color axis scaling according to cmin and cmax
            pass
        else:
            # use autoranging for color axis scale
            pass

    def _set_colormap(self, ax):
        """Set the colormap.""" 
        if DEBUG:
            print "Setting colormap"
        cmap = ax.getp('colormap')
        # cmap is plotting package dependent

    def _set_view(self, ax):
        """Set viewpoint specification."""
        if DEBUG:
            print "Setting view"
        cam = ax.getp('camera')
        view = cam.getp('view')
        
        width = self._g.width
        height = self._g.height
        self._g('resolution = %s;' % width)
        self._g('aspect = %f;' % ((height)/width))

        if view == 2:
            # setup a default 2D view
            #self._g('camtarget = collected;')
            #self._g('campos = Direction(0, 90, 10);')
            #self._g('camup = [0,1,0];')
            #self._g('camva = 30;')
            #self._g('camproj = 0;')
            self._g('camera = AutoCamera(collected);')
        elif view == 3:
            az = cam.getp('azimuth')
            el = cam.getp('elevation')
            if az is None or el is None:
                # azimuth or elevation is not given. Set up a default
                # 3D view (az=-37.5 and el=30 is the default 3D view in
                # Matlab).
                self._g('campos = Direction(-37.5, 30, 10);')
            else:
                # set a 3D view according to az and el
                self._g('campos = Direction(%s, %s, 10);' % (az,el))
            
            if cam.getp('cammode') == 'manual':
                # for advanced camera handling:
                roll = cam.getp('camroll')
                zoom = cam.getp('camzoom')
                dolly = cam.getp('camdolly')
                target = cam.getp('camtarget')
                position = cam.getp('campos')
                assert target != position, \
                       'camera target and position cannot be equal'
                up_vector = cam.getp('camup')
                view_angle = cam.getp('camva')
                projection = cam.getp('camproj')
                self._g('camtarget = [%s,%s,%s];' % target)
                self._g('campos = [%s,%s,%s];' % position)
                self._g('camup = [%s,%s,%s];' % up_vector)
                if view_angle is None:
                    view_angle = 30;
                self._g('camva = %s;' % view_angle)
                if projection == 'perspective':
                    self._g('camproj = 1;')
                else:
                    self._g('camproj = 0;')
            else:
                # set up some default values:
                self._g('camtarget = collected;')
                self._g('camproj = 0;')
                self._g('camup = [0,0,1];')
                self._g('camva = 30;')

            #self._g('camera = Camera(to=camtarget, from=campos, ' +
            #        'resolution=resolution, aspect=aspect, up=camup, ' +
            #        'angle=camva, ' +
            #        'perspective=camproj, background="black");')

            # use AutoCamera instead:
            self._g('camera = AutoCamera(camtarget, "off-bottom");')

    def _set_axis_props(self, ax):
        if DEBUG:
            print "Setting axis properties"
        self._set_title(ax)
        self._set_scale(ax)
        self._set_limits(ax)
        self._set_position(ax)
        self._set_axis_method(ax)
        self._set_daspect(ax)
        self._set_coordinate_system(ax)
        self._set_hidden_line_removal(ax)
        self._set_colorbar(ax)
        self._set_caxis(ax)
        self._set_colormap(ax)

        self._g('iact_mode = CollectNamed(2, "mode");')  # 2 = ROTATE
        hw_render_mode = 'opengl'
        hw_render_approx = 'none'
        if hw_render_mode == "none":
            options = '"rendering mode", "software"'
        else:
            options = '"rendering mode", "hardware", ' + \
                      ('"rendering approximation", "%s"' % hw_render_approx)
        options = options + ', "interaction mode", iact_mode'
        self._g('collected = Options(collected, %s);' % options)
        
        self._set_view(ax)
                        
        if ax.getp('visible'):
            self._set_labels(ax)
            self._set_box(ax)
            self._set_grid(ax)
            self._g('objectwithaxes = AutoAxes(collected, camera, labels, ' +
                    'frame=show_box, corners=collected, grid=show_grid, ' +
                    'colors={"red","yellow","yellow"}, ' +
                    'annotation={"grid","labels","ticks"});')
            #self._g('objectwithaxes = Scale(objectwithaxes, daspect);')
            self._g('camera = UpdateCamera(camera,objectwithaxes);')
            self._g('renderable = objectwithaxes;')
        else:
            # turn off all axis labeling, tickmarks, and background
            self._g('renderable = collected;')

    def _create_2D_vector_field(x, y, u, v):
        pass

    def _create_3D_vector_field(x, y, z, u, v, w):
        dataf = "x y z u v w"
        """file = /tmp/vector3D.dat
points = 25
format = ascii
interleaving = field
field = locations, field0
structure = 3-vector, 3-vector
type = float, float

end
"""
        pass

    def _create_2D_scalar_data_file(self, x, y, z,
                                    regular_grid=False, indexing='ij'):
        tmp = tempfile.mktemp()
        # first create data file:
        data_file = open(tmp+'.dat', 'w')
        nx, ny = shape(z)
        if regular_grid:
            for i in range(nx):
                for j in range(ny):
                    data_file.write("%5.3lf\t" % z[i,j])
                data_file.write("\n")
        else:
            if shape(x) != (nx,ny) and shape(y) != (nx,ny):
                x, y = ndgrid(x,y,sparse=False)
            for i in range(nx):
                for j in range(ny):
                    data_file.write("%5.3lf\t%5.3lf\t%5.3lf\n" % \
                                    (x[i,j],y[i,j],z[i,j]))
        data_file.write("\n")
        data_file.close()
        # then create general header file:
        header_file = open(tmp+'.general', 'w')
        if regular_grid:
            x0 = x[0,0]
            y0 = y[0,0]
            if indexing == 'ij':
                dx = x[1,0] - x[0,0]
                dy = y[0,1] - y[0,0]
            else:
                dx = x[0,1] - x[0,0]
                dy = y[1,0] - y[0,0]
            header_file.write("""file = %s
grid = %d x %d
format = ascii
interleaving = record
majority = row
field = field0
structure = scalar
type = float
dependency = positions
positions = regular, regular, %s, %s, %s, %s

end
""" % (data_file.name,nx,ny,x0,dx,y0,dy))
        else:
            header_file.write("""file = %s
grid = %d x %d
format = ascii
interleaving = field
majority = row
field = locations, field0
structure = 2-vector, scalar
type = float, float

end
""" % (data_file.name,nx,ny))
        header_file.close()
        return header_file.name

    def _create_3D_scalar_data_file(self, x, y, z, v,
                                    regular_grid=False, indexing='ij'):
        tmp = tempfile.mktemp()
        # create data file:
        data_file = open(tmp+'.dat', 'w')
        nx, ny, nz = shape(v)
        if regular_grid:
            for i in range(nx):
                for j in range(ny):
                    for k in range(nz):
                        data_file.write("%5.3lf\t" % v[i,j,k])
                    data_file.write("\n")
        else:
            if shape(x) != (nx,ny,nz) and shape(y) != (nx,ny,nz) \
                   and shape(z) != (nx,ny,nz):
                x, y, z = ndgrid(x,y,z,sparse=False)
            for i in range(nx):
                for j in range(ny):
                    for k in range(nz):
                        data_file.write("%5.3lf\t%5.3lf\t%5.3lf\t%5.3lf\n" % \
                                        (x[i,j,k],y[i,j,k],z[i,j,k],v[i,j,k]))
        data_file.write("\n")
        data_file.close()
        # create general header file:
        header_file = open(tmp+'.general', 'w')
        if regular_grid:
            x0 = x[0,0,0]
            y0 = y[0,0,0]
            z0 = z[0,0,0]
            if indexing == 'ij':
                dx = x[1,0,0] - x[0,0,0]
                dy = y[0,1,0] - y[0,0,0]
                dz = z[0,0,1] - z[0,0,0]
            else:
                dx = x[0,1,0] - x[0,0,0]
                dy = y[1,0,0] - y[0,0,0]
                dz = z[0,0,1] - z[0,0,0]
            header_file.write("""file = %s
grid = %d x %d x %d
format = ascii
interleaving = record
majority = row
field = field0
structure = scalar
type = float
dependency = positions
positions = regular, regular, regular, %s, %s, %s, %s, %s, %s

end
""" % (data_file.name,nx,ny,nz,x0,dx,y0,dy,z0,dz))
        else:
            header_file.write("""file = %s
grid = %d x %d x %d
format = ascii
interleaving = field
majority = row
field = locations, field0
structure = 3-vector, scalar
type = float, float

end
""" % (data_file.name,nx,ny,nz))
        header_file.close()
        return header_file.name

    def _create_2D_scalar_field(self, x, y, z, id,
                                regular_grid=False,
                                indexing='ij'):
        nx, ny = shape(z)
        if shape(x) != (nx,ny) and shape(y) != (nx,ny):
            x, y = ndgrid(x,y,sparse=False,indexing=indexing)

        # the scalar field should be a string on the form
        # 'z0 z1 z2 ... zn' where n=nx*ny*nz: 
        z = ravel(z).tolist()
        scalar_field = ' '.join([str(i) for i in z])

        data = 'data%s' % id
        if regular_grid:
            x0 = x[0,0]
            y0 = y[0,0]
            if indexing == 'ij':
                dx = x[1,0] - x0
                dy = y[0,1] - y0
            else:
                dx = x[0,1] - x0
                dy = y[1,0] - y0

            self._g('%s = Construct([%s %s],[%s %s],[%s %s],{%s});' % \
                    (data,x0,y0,dx,dy,nx,ny,scalar_field))
        else:
            x = ravel(x).tolist()
            y = ravel(y).tolist()
            # create a string with grid positions on the form
            # '[x0 y0][x1 y0] ... [xn y0][x0 y1] ... [xn yn]':
            positions = ''.join(['[%s %s]' % (xp,yp) for xp, yp in zip(x,y)])

            self._g('%s = Construct({%s},NULL,[%d %d],{%s});' % \
                    (data,positions,nx,ny,scalar_field))
        return data

    def _create_3D_scalar_field(self, x, y, z, v, id,
                                regular_grid=False,
                                indexing='ij'):
        nx, ny, nz = shape(v)
        if shape(x) != (nx,ny,nz) and shape(y) != (nx,ny,nz) \
               and shape(z) != (nx,ny,nz):
            x, y, z = ndgrid(x,y,z,sparse=False,indexing=indexing)

        # the scalar field should be a string on the form
        # 'z0 z1 z2 ... zn' where n=nx*ny*nz: 
        v = ravel(v).tolist()
        scalar_field = ' '.join([str(i) for i in v])

        data = 'data%s' % id
        if regular_grid:
            x0 = x[0,0,0]
            y0 = y[0,0,0]
            z0 = z[0,0,0]
            if indexing == 'ij':
                dx = x[1,0,0] - x0
                dy = y[0,1,0] - y0
                dz = z[0,0,1] - z0
            else:
                dx = x[0,1,0] - x0
                dy = y[1,0,0] - y0
                dz = z[0,0,1] - z0
            self._g('%s = Construct([%s %s %s],[%s %s %s],[%s %s %s],{%s});' \
                    % (data,x0,y0,z0,dx,dy,dz,nx,ny,nz,scalar_field))
        else:
            x = ravel(x).tolist()
            y = ravel(y).tolist()
            z = ravel(z).tolist()
            # create a string with grid positions on the form
            # '[x0 y0 z0][x1 y0 z0] ... [xn y0 z0][x0 y1 z0] ... [xn yn zn]'
            positions = ''.join(['[%s %s %s]' % (xp,yp,zp) \
                                 for xp, yp, zp in zip(x,y,z)])

            self._g('%s = Construct({%s},NULL,[%d %d %d],{%s});' % \
                    (data,positions,nx,ny,nz,scalar_field))
        return data

    def _get_linespecs(self, item):
        """
        Return the line marker, line color, line style, and
        line width of the item.
        """
        
        marker = self._markers[item.getp('linemarker')]
        color = self._colors[item.getp('linecolor')]
        style = self._line_styles[item.getp('linetype')]
        width = item.getp('linewidth')
        return marker, color, style, width

    def _add_line(self, item):
        """Add a 2D or 3D curve to the scene."""
        if DEBUG:
            print "Adding a line"
        # get data:
        x = item.getp('xdata')
        y = item.getp('ydata')
        z = item.getp('zdata')
        # get line specifiactions:
        marker, color, style, width = self._get_linespecs(item)

        if z is not None:
            # zdata is given, add a 3D curve:
            pass
        else:
            # no zdata, add a 2D curve:
            pass

    def _add_surface(self, item, id, shading='faceted'):
        if DEBUG:
            print "Adding a surface"
        x = item.getp('xdata')  # grid component in x-direction
        y = item.getp('ydata')  # grid component in y-direction
        z = item.getp('zdata')  # scalar field
        c = item.getp('cdata')  # pseudocolor data (can be None)
        indexing = item.getp('indexing')
        
        #general_file = self._create_2D_scalar_data_file(x, y, z,
        #                                                regular_grid=False,
        #                                                indexing=indexing)
        #self._g('imported%s = Import("%s",format="general");' \
        #        % (id,general_file))
        #data_field = 'imported%s' % id
        data_field = self._create_2D_scalar_field(x, y, z, id,
                                                  regular_grid=False,
                                                  indexing=indexing)
        self._g('colored%s = AutoColor(%s);' % (id,data_field))
        self._g('rubbersheet%s = RubberSheet(colored%s,scale=1);' % (id,id))
        dar = self._ax.getp('daspect')
        self._g('rubbersheet%s = Scale(rubbersheet%s,[%s %s %s]);' % \
                (id,id,dar[0],dar[1],dar[2]))
                
        contours = item.getp('contours')
        if contours:
            # the current item is produced by meshc or surfc and we
            # should therefore add contours at the bottom:
            self._add_contours(contours, id, placement='bottom')

        if item.getp('wireframe'):
            # wireframe mesh (as produced by mesh or meshc)
            self._g('obj%s = ShowConnections(rubbersheet%s);' % (id,id))
        else:
            # colored surface (as produced by surf, surfc, or pcolor)
            # use keyword argument shading to set the color shading mode
            pcolor = item.getp('function') == 'pcolor'
            if pcolor:
                self._g('rubbersheet%s = colored%s;' % (id,id))
            if shading == 'flat':
                self._g('obj%s = FaceNormals(rubbersheet%s);' % (id,id))
            elif shading == 'interp':
                self._g('obj%s = Normals(rubbersheet%s);' % (id,id))
                #self._g('obj%s = rubbersheet%s;' % (id,id))
            else:
                self._g('colored_mesh%s = Color(%s,"black");' % \
                        (id,data_field)) # FIXME: add sup. for other colors
                self._g(('mesh_rubbersheet%s = RubberSheet(colored_mesh%s, ' +\
                        'scale=1);') % (id,id))
                self._g(('mesh_rubbersheet%s = Scale(mesh_rubbersheet%s, ' + \
                         '[%s %s %s]);') % (id,id,dar[0],dar[1],dar[2]))
                if pcolor:
                    self._g('mesh_rubbersheet%s = colored_mesh%s;' % \
                            (id,id))
                self._g('faceted%s = ShowConnections(mesh_rubbersheet%s);' % \
                        (id,id))
                self._g('collected = Append(collected,faceted%s);' % id)
                self._g('obj%s = FaceNormals(rubbersheet%s);' % (id,id))

        self._g('collected = Append(collected, obj%s);' % id)

    def _add_contours(self, item, id, placement=None):
        # The placement keyword can be either None or 'bottom'. The
        # latter specifies that the contours should be placed at the
        # bottom (as in meshc or surfc).
        if DEBUG:
            print "Adding contours"
        x = item.getp('xdata')  # grid component in x-direction
        y = item.getp('ydata')  # grid component in y-direction
        z = item.getp('zdata')  # scalar field
        indexing = item.getp('indexing')
        dar = self._ax.getp('daspect')

        #general_file = self._create_2D_scalar_data_file(x, y, z)
        #self._g('imported%s = Import("%s",format="general");' \
        #        % (id,general_file))
        data_field = self._create_2D_scalar_field(x, y, z, id,
                                                  indexing=indexing)
        self._g('colored%s = AutoColor(%s);' % (id,data_field))

        filled = item.getp('filled')  # draw filled contour plot if True

        cvector = item.getp('cvector')
        clevels = item.getp('clevels')  # number of contour levels
        if cvector is None:
            # the contour levels are chosen automatically
            self._g('cvector = NULL;')
        else:
            cvector = '{' + ','.join([str(c) for c in cvector]) + '}'
            self._g('cvector = %s;' % cvector)
        self._g('clevels = %s;' % clevels)
        
        location = item.getp('clocation')
        if location == 'surface':
            # place the contours at the corresponding z level (contour3)
            self._g('rubbersheet%s = RubberSheet(colored%s,scale=1);' % \
                    (id,id))
            self._g('rubbersheet%s = Scale(rubbersheet%s, [%s %s %s]);' % \
                    (id,id,dar[0],dar[1],dar[2]))
            self._g(('obj%s = Isosurface(rubbersheet%s, value=cvector, ' + \
                     'number=clevels);') % (id,id))
        elif location == 'base':
            if placement == 'bottom':
                # place the contours at the bottom (as in meshc or surfc)
                pass
            else:
                # standard contour plot
                pass
            self._g(('obj%s = Isosurface(colored%s, value=cvector, ' + \
                     'number=clevels);') % (id,id))
            self._g('obj%s = Scale(obj%s, [%s %s %s]);' % \
                    (id,id,dar[0],dar[1],dar[2]))

        if item.getp('clabels'):
            # add labels on the contour curves
            pass

        linewidth = item.getp('linewidth')
        if linewidth:
            self._g('obj%s = Options(obj%s, "line width", %s);' % \
                    (id,id,linewidth))
        self._g('collected = Append(collected, obj%s);' % id)
    
    def _add_vectors(self, item):
        if DEBUG:
            print "Adding vectors"
        # uncomment the following command if there is no support for
        # automatic scaling of vectors in the current plotting package:
        #item.scale_vectors()

        # grid components:
        x, y, z = item.getp('xdata'), item.getp('ydata'), item.getp('zdata')
        # vector components:
        u, v, w = item.getp('udata'), item.getp('vdata'), item.getp('wdata')
        # get line specifiactions (marker='.' means no marker):
        marker, color, style, width = self._get_linespecs(item)

        # scale the vectors according to this variable (scale=0 should
        # turn off automatic scaling):
        scale = item.getp('arrowscale')

        filled = item.getp('filledarrows') # draw filled arrows if True

        if z is not None and w is not None:
            # draw velocity vectors as arrows with components (u,v,w) at
            # points (x,y,z):
            pass
        else:
            # draw velocity vectors as arrows with components (u,v) at
            # points (x,y):
            pass

    def _add_streams(self, item):
        if DEBUG:
            print "Adding streams"
        # grid components:
        x, y, z = item.getp('xdata'), item.getp('ydata'), item.getp('zdata')
        # vector components:
        u, v, w = item.getp('udata'), item.getp('vdata'), item.getp('wdata')
        # starting positions for streams:
        sx, sy, sz = item.getp('startx'), item.getp('starty'), item.getp('startz')

        if item.getp('tubes'):
            # draw stream tubes from vector data (u,v,w) at points (x,y,z)
            n = item.getp('n') # no points along the circumference of the tube
            scale = item.getp('tubescale')
            pass
        elif item.getp('ribbons'):
            # draw stream ribbons from vector data (u,v,w) at points (x,y,z)
            width = item.getp('ribbonwidth')
            pass
        else:
            if z is not None and w is not None:
                # draw stream lines from vector data (u,v,w) at points (x,y,z)
                pass
            else:
                # draw stream lines from vector data (u,v) at points (x,y)
                pass
            pass

    def _add_isosurface(self, item, id):
        if DEBUG:
            print "Adding a isosurface"
        # grid components:
        x, y, z = item.getp('xdata'), item.getp('ydata'), item.getp('zdata')
        v = item.getp('vdata')  # volume
        c = item.getp('cdata')  # pseudocolor data
        isovalue = item.getp('isovalue')
        indexing = item.getp('indexing')
        dar = self._ax.getp('daspect')
        
        #general_file = self._create_3D_scalar_data_file(x, y, z, v)
        #self._g('data%s = Import("%s",format="general");' \
        #        % (id,general_file))
        data_field = self._create_3D_scalar_field(x, y, z, v, id,
                                                  regular_grid=False,
                                                  indexing=indexing)
        self._g('colored%s = AutoColor(%s);' % (id,data_field))
        self._g('obj%s = Isosurface(colored%s, value=%s);' % \
                (id,id,isovalue))
        self._g('obj%s = Scale(obj%s,[%s %s %s]);' % \
                (id,id,dar[0],dar[1],dar[2]))

        self._g('collected = Append(collected, obj%s);' % id)
    
    def _add_slices(self, item, id, shading='faceted'):
        if DEBUG:
            print "Adding slices in a volume"
        # grid components:
        x, y, z = item.getp('xdata'), item.getp('ydata'), item.getp('zdata')
        v = item.getp('vdata')  # volume
        indexing = item.getp('indexing')
        dar = self._ax.getp('daspect')
        xmin, xmax, ymin, ymax, zmin, zmax = item.get_limits()
        center = [(xmax+xmin)/2, (ymax+ymin)/2, (zmax+zmin)/2]

        data_field = self._create_3D_scalar_field(x, y, z, v, id,
                                                  regular_grid=False,
                                                  indexing=indexing)
        self._g('colored%s = AutoColor(%s);' % (id,data_field))
        
        self._g('slices%s = Collect();' % id)
        
        sx, sy, sz = item.getp('slices')
        if rank(sz) == 2:
            # sx, sy, and sz defines a surface
            pass
        else:
            # sx, sy, and sz is either numbers or vectors with numbers
            points = []
            normals = []
            sx = ravel(sx)#/dar[0]
            sy = ravel(sy)#/dar[1]
            sz = ravel(sz)#/dar[2]
            for i in range(len(sx)):
                normals.append([1,0,0])
                points.append([sx[i], center[1], center[2]])
            for i in range(len(sy)):
                normals.append([0,1,0])
                points.append([center[0], sy[i], center[2]])
            for i in range(len(sz)):
                normals.append([0,0,1])
                points.append([center[0], center[1], sz[i]])
            for i in range(len(normals)):
                normal = normals[i]
                point = points[i]
                self._g('normal = [%s %s %s];' % \
                        (normal[0],normal[1],normal[2]))
                self._g('point = [%s %s %s];' % (point[0],point[1],point[2]))
                self._g('slice = MapToPlane(colored%s, point, normal);' % id)
                if shading == 'interp':
                    self._g('slice = Normals(slice);')
                elif shading == 'flat':
                    self._g('slice = FaceNormals(slice);')
                else:
                    self._g('mesh_slice%s = Color(slice,"black");' % id)
                    self._g('mesh_slice%s = ShowConnections(mesh_slice%s);' % \
                            (id,id))
                    self._g('slices%s = Append(slices%s, mesh_slice%s);' % \
                            (id,id,id))
                #self._g(
                self._g('slices%s = Append(slices%s, slice);' % (id,id))

        self._g('collected = Append(collected, slices%s);' % id)

    def _add_contourslices(self, item):
        if DEBUG:
            print "Adding contours in slice planes"
        # grid components:
        x, y, z = item.getp('xdata'), item.getp('ydata'), item.getp('zdata')
        v = item.getp('vdata')  # volume

        sx, sy, sz = item.getp('slices')
        if rank(sz) == 2:
            # sx, sy, and sz defines a surface
            pass
        else:
            # sx, sy, and sz is either numbers or vectors with numbers
            pass

        cvector = item.getp('cvector')
        clevels = item.getp('clevels')  # number of contour levels per plane
        if cvector is None:
            # the contour levels are chosen automatically
            #cvector =
            pass
        pass

    def _set_figure_size(self, fig):
        if DEBUG:
            print "Setting figure size"
        width, height = fig.getp('size')
        if width and height:
            # set figure width and height
            self._g.root.geometry('%sx%s' % (width,height))
            self._g.root.update()
        else:
            # use the default width and height in plotting package
            pass

    def figure(self, *args, **kwargs):
        # Extension of BaseClass.figure:
        # add a plotting package figure instance as fig._g and create a
        # link to it as self._g
        fig = BaseClass.figure(self, *args, **kwargs)
        try:
            fig._g
        except:
            # create plotting package figure and save figure instance
            # as fig._g
            name = 'Figure ' + str(fig.getp('number'))
            if DEBUG:
                print "creating figure %s in backend" % name

            fig._g = _DXFigure(self, title=name)
            
        self._g = fig._g  # link for faster access
        return fig

    def _fix_latex(self, legend):
        """Remove latex syntax a la $, \, {, } etc."""
        legend = legend.strip()
        # General fix of latex syntax (more readable)
        legend = legend.replace('**', '^')  
        #legend = legend.replace('*', '')
        legend = legend.replace('$', '')
        legend = legend.replace('{', '')
        legend = legend.replace('}', '')
        legend = legend.replace('\\', '')
        return legend
        
    def _replot(self):
        """Replot all axes and all plotitems in the backend."""
        # NOTE: only the current figure (gcf) is redrawn.
        if DEBUG:
            print "Doing replot in backend"
        
        fig = self.gcf()
        # reset the plotting package instance in fig._g now if needed
        fig._g.reset()
        
        # include some useful macros:
        # (look at DXLLoadMacroFile)
        fig._g('//include "AutoScaleMacro.net"')
        fig._g('//include "ArrangeMemberMacro.net"')
        fig._g('//include "CappedIsoMacro.net"')
        
        self._set_figure_size(fig)

        #DX.exDXLBeginMacroDefinition(self._g.conn, 'macro main()')
                
        width = self._g.width
        height = self._g.height
        parent = self._g.frame.winfo_id()
        depth = self._g.frame.winfo_depth()
        self._g(('where, size, events = SuperviseWindow("Easyviz", ' +
                 'display=NULL, size=[%d,%d], offset=NULL, ' +
                 'parent=%s, depth=%d);') % (width, height, parent, depth))
        
        nrows, ncolumns = fig.getp('axshape')
        for axnr, ax in fig.getp('axes').items():
            self._ax = ax          # create link for easier access later
            self._axnr = axnr - 1  # same
            pth = ax.getp('pth')
            if pth:
                # create axes in tiled position
                # this is subplot(nrows,ncolumns,axnr)
                pass
            i = 0
            self._g('collected = Collect();')
            plotitems = ax.getp('plotitems')
            plotitems.sort(self._cmpPlotProperties)
            for item in plotitems:
                item_id = str(i)
                func = item.getp('function') # function that produced this item
                if isinstance(item, Line):
                    self._add_line(item)
                elif isinstance(item, Surface):
                    self._add_surface(item, item_id, shading=ax.getp('shading'))
                elif isinstance(item, Contours):
                    self._add_contours(item, item_id)
                elif isinstance(item, VelocityVectors):
                    self._add_vectors(item)
                elif isinstance(item, Streams):
                    self._add_streams(item)
                elif isinstance(item, Volume):
                    if func == 'isosurface':
                        self._add_isosurface(item, item_id)
                    elif func == 'slice_':
                        self._add_slices(item, item_id,
                                         shading=ax.getp('shading'))
                    elif func == 'contourslice':
                        self._add_contourslices(item)
                legend = self._fix_latex(item.getp('legend'))
                if legend:
                    # add legend to plot
                    pass
                i += 1

            self._rendermode = "hardware"  # hardware or software rendering
            self._interactionmode = 0      # 0:rotate, 1:pan, 2:zoom

            self._set_axis_props(ax)

            nr = self._axnr
            self._g(('object%s, cam%s, where%s = ArrangeMember(renderable, ' +
                     'renderMode="%s", defaultCamera=camera, ' + 
                     'interactionMode=%d, parentSize=size, parent=where, ' + 
                     'title="image", totalSubimages={%d}, ' + 
                     'nHorizontal={%d}, which={%d});') % \
                    (nr,nr,nr,self._rendermode,self._interactionmode,
                     nrows*ncolumns,ncolumns,nr))
            #self._g('Display(renderable, camera, where=where);')

        #DX.exDXLEndMacroDefinition(self._g.conn)
                    
        if self.getp('show'):
            # display plot on the screen
            if DEBUG:
                print "\nDumping plot data to screen\n"
                debug(self)
            pass
        self._g.display(show=self.getp('show'))

    def hardcopy(self, filename, **kwargs):
        """
        Supported extensions:

          '.ps'   (PostScript)
          '.epsf' (Encapsualted PostScript)
          '.gif'  (Graphics Interchange Format)
          '.tiff' (Tag Image File Format)
          '.miff' (Magick Image File Format)
          FIXME: Add more formats like rgb, r+g+b, yuv, ...

        Optional arguments:

          color       -- True (colors) or False (black and white).
          size        -- A tuple (width,height) to set the size in inches of
                         the image. The default is (8.5,11).
          dpi         -- Set the number of dots (pixels) per inch in the
                         hardcopy image. Use this only if you want to
                         explicitly set the number of dots per inch. 
          orientation -- 'auto' (default), 'portrait', or 'landscape'. Only
                         available for PostScript output.
          margin      -- Sets the desired margin around the image. The 
                         default is 0.5 inch.
          width       -- FIXME ...
          height      -- FIXME ...
          gamma       -- Sets the gamma correction factor for the output
                         image. The default is 2.0. Avaliable for all output
                         formats.
          delayed     -- If True this creates an image-with-colormap. Only
                         available for PostScript, TIFF, and MIFF formats.
                         The default is False.
          frame       -- FIXME ...

        """
        if DEBUG:
            print "Hardcopy to %s" % filename

        self.setp(**kwargs)
        color = self.getp('color')
        replot = kwargs.get('replot', True)
        if replot:
            self._replot()

        if color:
            color = 'color'
        else:
            color = 'gray'
        size = kwargs.get('size', (8.5,11))
        dpi = kwargs.get('dpi', None)
        orientation = kwargs.get('orientation', 'auto')
        margin = kwargs.get('margin', 0.5)
        width = kwargs.get('width', None)
        height = kwargs.get('height', None)
        gamma = kwargs.get('gamma', 2.0)
        delayed = kwargs.get('delayed', False)

        basename, ext = os.path.splitext(filename)
        if not ext:
            ext = '.ps'
            filename += ext

        format = ext[1:]
        options = ''
        if ext in ['.ps', '.epsf']:
            options += '%s page=%sx%s margin=%s orient=%s' % \
                       (color,size[0],size[1],margin,orientation)
            if dpi is not None:
                options += ' dpi=%s' % dpi
            if width is not None:
                options += ' width=%s' % width
            if height is not None:
                options += ' height=%s' % height
        if ext in ['.ps', '.epsf', '.tiff', '.miff']:
            if delayed:
                options += ' delayed=1'
        options += ' gamma=%s' % gamma

        format_str = '%s %s' % (format,options)
        # render image(s):
        nrows, ncolumns = self.gcf().getp('axshape')
        self._g('images = Collect();')
        for i in range(nrows*ncolumns):
            self._g('image%d = Render(object%d, cam%d);' % (i,i,i))
            self._g('images = Append(images, image%d);' % i)
        self._g('arranged = Arrange(images, %d);' % ncolumns)
        # write to file:
        self._g('WriteImage(arranged,"%s","%s");' % (basename,format_str))

    # reimplement methods like clf, closefig, closefigs
    def clf(self):
        self._g.reset()
        self._g.root.withdraw()
        BaseClass.clf(self)

    def save_script(self, filename):
        """
        Save the DX commands used for the plot in the current figure in a
        script file that can be run by the command dx -script <filename>.
        Note that some aspects of the script has to be modified in order
        for it to work as intended (like uncommenting the necessary macros
        and removing the parent option in the SuperviseWindow.
        """
        f = open(filename, 'w')
        f.write(self._g.script)
        f.close()

    # implement colormap functions here
    #def jet(self, m=None):
    #    """Variant of hsv."""
    #    pass


    # Now we add the doc string from the methods in BaseClass to the
    # methods that are reimplemented in this backend:
    for cmd in BaseClass._matlab_like_cmds:
        if not '__' in cmd and hasattr(BaseClass, cmd):
            m1 = eval('BaseClass.%s' % cmd)
            try:
                m2 = eval('%s' % cmd)
            except NameError:
                pass
            else:
                if m1.__doc__ != m2.__doc__:
                    if m2.__doc__ is None:
                        m2.__doc__ = ""
                    m2.__doc__ = m1.__doc__ + m2.__doc__
    

plt = DXBackend()    # create backend instance
use(plt, globals())  # export public namespace of plt to globals()
backend = os.path.splitext(os.path.basename(__file__))[0][:-1]



"""
Python synthax higlighting

Borrowed from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52298
"""

# Original copyright : 2001 by Juergen Hermann <jh@web.de>
import cgi, cStringIO, keyword, token, tokenize

class PythonParser:
    """ Send colored python source.
    """

    _KEYWORD = token.NT_OFFSET + 1
    _TEXT    = token.NT_OFFSET + 2

    def __init__(self):
        """ Store the source text.
        """
        self._tags = {
            token.NUMBER: 'pynumber',
            token.OP: 'pyoperator',
            token.STRING: 'pystring',
            tokenize.COMMENT: 'pycomment',
            tokenize.ERRORTOKEN: 'pyerror',
            self._KEYWORD: 'pykeyword',
            self._TEXT: 'pytext',
        }
        self.pysrcid = 0;


    def __call__(self, raw):
        """ Parse and send the colored source.
        """
        self.out = cStringIO.StringIO()
        self.raw = raw.expandtabs().strip()
        # store line offsets in self.lines
        self.lines = [0, 0]
        pos = 0
        while 1:
            pos = self.raw.find('\n', pos) + 1
            if not pos: break
            self.lines.append(pos)
        self.lines.append(len(self.raw))
        #
        # parse the source and write it
        self.pos = 0
        text = cStringIO.StringIO(self.raw)
        self.out.write("<table width=100% cellpadding=0 cellspacing=0 " +
                     """onclick="toggle_hidden('pysrc%d','toggle%d');"><tr>
                        <td rowspan="3"> """ % (self.pysrcid, self.pysrcid) )
        self.out.write("""<div class="pysrc" id="pysrc%dinv" style="display:
                       none">...</div>"""% self.pysrcid)
        self.out.write('<div class="pysrc" id="pysrc%d" style="display: block ">'% self.pysrcid)

        try:
            tokenize.tokenize(text.readline, self.format)
        except tokenize.TokenError, ex:
            msg = ex[0]
            line = ex[1][0]
            print >> self.out, ("<h3>ERROR: %s</h3>%s" %
                (msg, self.raw[self.lines[line]:]))
        self.out.write('</div>')
        self.out.write('''
                       </td> 
                       <td colspan="2" class="collapse bracket"></td>
                       </tr>
                       <tr>
                       <td class="bracketfill"></td>
                       <td width=5px class="collapse"> 
                           <div id="toggle%d">
                           <small>.</small>
                           </div>
                       </td>
                       </tr>
                       <tr><td colspan="2" class="collapse bracket"></td>
                       </tr>
                       </table>
                       ''' % (self.pysrcid))
        self.pysrcid += 1
        return self.out.getvalue()

    def format(self, toktype, toktext, (srow, scol), (erow, ecol), line):
        """ Token handler.
        """
        
        # calculate new positions
        oldpos = self.pos
        newpos = self.lines[srow] + scol
        self.pos = newpos + len(toktext)
        #
        # handle newlines
        if toktype in [token.NEWLINE, tokenize.NL]:
            # No need, for that: the css attribute "white-space: pre;" will 
            # take care of that.
            self.out.write("\n")
            return
        #
        # send the original whitespace, if needed
        if newpos > oldpos:
            self.out.write(self.raw[oldpos:newpos])
        #
        # skip indenting tokens
        if toktype in [token.INDENT, token.DEDENT]:
            self.pos = newpos
            return
        #
        # map token type to a color group
        if token.LPAR <= toktype and toktype <= token.OP:
            toktype = token.OP
        elif toktype == token.NAME and keyword.iskeyword(toktext):
            toktype = self._KEYWORD
        style = self._tags.get(toktype, self._tags[self._TEXT])
        #
        # send text
        self.out.write('<span class="%s">' % (style, ))
        self.out.write(cgi.escape(toktext))
        self.out.write('</span>')

python2html = PythonParser()





#!/usr/bin/env python

from scitools.std import *
from numpy.random import rand, seed
seed(12)  # fix to ensure same results each time this program is run

# simple example:
data = rand(4,3)
print data
bar(data)
title('Bars from a matrix: 4 groups of 3 bars each')
# alternative:
figure()
bar(data, 
    barticks=['group 1', 'group 2', 'group 3', 'group 4'],
    legend=['bar 1', 'bar 2', 'bar 3'],
    axis=[-1, data.shape[0], 0, 1.3],
    ylabel='Normalized CPU time',
    title='Bars from a matrix, now with more annotations')
    

# example with a vector:
figure()
x = linspace(-3,3,15)
y = exp(-x**2)
if backend in ['gnuplot', 'matplotlib']:
    w = 7   # increase width of bars
else:
    w = 0.8 # default
bar(y, width=w)
title('Bars at 15 points for the exp(-x**2) functions')

# draw a bar graph from a dicitonary:
figure()
data = {'method1': {'thing1': 3, 'thing2': 2, 'thing3': 4},
        'method2': {'thing1': 4, 'thing2': 4.5, 'thing3': 2},
        'method3': {'thing1': 1, 'thing2': 4, 'thing3': 2},}
bar(data, 
    rotated_barticks=True,
    hardcopy='tmp_bar1a.eps', color=True,
    ylabel='response',
    axis=[-1, len(data), 0, 5],
    title='Bars from nested dictionary')

# red bars with borders turned off:
figure()
bar(rand(5,4), 'r', shading='interp', 
    title='Red bars with borders turned off')

# specify face and edge color:
figure()
bar(rand(3,3), facecolor='b', edgecolor='r',
    title='Bars with specified face and edge colors')

raw_input('Press Return key to quit: ')



#!/usr/bin/env python
"""Install SciTools
Usage:

python setup.py install [, --prefix=...]
"""

__author__ = 'Hans Petter Langtangen <hpl@simula.no>, Rolv E. Bredesen <rolv@simula.no>'

from distutils.core import setup
import os, glob

# this script applies to a directory tree with the SciTools++ suite
external_packages = ['preprocess', 'epydoc', 'Gnuplot', 'IPython',
                     'Pmw', 'Scientific']
internal_packages = ['doconce',
                     'scitools', 'scitools.easyviz', 'scitools.pyPDE',
                     ]

setup(
    version = "1.0",
    author = "Rolv Bredesen, Hans Petter Langtangen",
    author_email = "<hpl@simula.no>",
    description = __doc__,
    license = "Well",
    name = "SciTools++",
    url = "",
    package_dir = {'': 'lib'},
    # packages to install:
    packages = internal_packages + external_packages,
    # standalone scripts :
    scripts = [os.path.join('bin', f) \
               for f in os.listdir('bin') if not f.startswith('.')],
    )
    
    



from scitools.std import *   # for curve plotting

def f(t):
    return t**2*exp(-t**2)

t = linspace(0, 3, 51)    # 51 points between 0 and 3
y = f(t)
plot(t, y,
     xlabel='t',
     ylabel='y',
     legend='t^2*exp(-t^2)',
     axis=[0, 3, -0.05, 0.6],
     title='My First Easyviz Demo',
     savefig='tmp1.eps',  # or hardcopy='tmp1.eps'
     show=True)

raw_input('Press Return key to quit: ')



#!/usr/bin/env python
"""
Class for uniform and non-uniform grid on an interval, rectangle, or box.
"""

from scitools.errorcheck import right_type, wrong_type
from scitools.numpyutils import ndgrid, ndarray, wrap2callable, array, \
     zeros, linspace

# constants for indexing the space directions:
X = X1 = 0
Y = X2 = 1
Z = X3 = 2


class UniformBoxGrid(object):
    """
    Simple uniform grid on an interval, rectangle, box, or hypercube.

    =============      ====================================================
      Attributes                           Description
    =============      ====================================================
    nsd                no of spatial dimensions in the grid
    min_coor           array of minimum coordinates
    max_coor           array of maximum coordinates
    division           array of cell divisions in the 
    delta              array of grid spacings
    dirnames           names of the space directions ('x', 'y', etc.)
    shape              (nx+1, ny+1, ...); dimension of array over grid
    coor               list of coordinates; self.coor[Y][j] is the
                       the j-th coordinate in direction Y (=1)
                       X, Y, Z are predefined constants 0, 1, 2
    coorv              expanded version of coor for vectorized expressions
                       (in 2D, self.coorv[0] = self.coor[0][:,newaxis])
    tolerance          small geometric tolerance based on grid coordinates
    npoints            total number of grid points
    =============      ====================================================

    """
    def __init__(self,
                 min=(0,0),                  # minimum coordinates
                 max=(1,1),                  # maximum coordinates
                 division=(4,4),             # cell divisions
                 dirnames=('x', 'y', 'z')):  # names of the directions
        """
        Initialize a BoxGrid by giving domain range (minimum and
        maximum coordinates: min and max tuples/lists/arrays)
        and number of cells in each space direction (division tuple/list/array).
        The dirnames tuple/list holds the names of the coordinates in
        the various spatial directions.

        >>> g = UniformBoxGrid(min=0, max=1, division=10)
        >>> g = UniformBoxGrid(min=(0,-1), max=(1,1), division=(10,4))
        >>> g = UniformBoxGrid(min=(0,0,-1), max=(2,1,1), division=(2,3,5))
        """
        # Allow int/float specifications in one-dimensional grids
        # (just turn to lists for later multi-dimensional processing)
        if isinstance(min, (int,float)):
            min = [min]
        if isinstance(max, (int,float)):
            max = [max]
        if isinstance(division, (int,float)):
            division = [division]
        if isinstance(dirnames, str):
            dirnames = [dirnames]
        
        self.nsd = len(min)
        # strip dirnames down to right space dim (in case the default
        # with three components were unchanged by the user):
        dirnames = dirnames[:self.nsd]

        # check consistent lengths:
        for a in max, division:
            if len(a) != self.nsd:
                raise ValueError(
                    'Incompatible lengths of arguments to constructor'\
                    ' (%d != %d)' % (len(a), self.nsd))

        self.min_coor = array(min, float)
        self.max_coor = array(max, float)
        self.dirnames = dirnames
        self.division = division
        self.coor = [None]*self.nsd
        self.shape = [0]*self.nsd
        self.delta = zeros(self.nsd)

        for i in range(self.nsd):
            self.delta[i] = \
                 (self.max_coor[i] -  self.min_coor[i])/float(self.division[i])
            self.shape[i] = self.division[i] + 1  # no of grid points
            self.coor[i] = \
                 linspace(self.min_coor[i], self.max_coor[i], self.shape[i])
        self._more_init()

    def _more_init(self):
        self.shape = tuple(self.shape)
        self.coorv = ndgrid(*self.coor)
        if not isinstance(self.coorv, (list,tuple)):
            # 1D grid, wrap self.coorv as list:
            self.coorv = [self.coorv]

        self.npoints = 1
        for i in range(len(self.shape)):
            self.npoints *= self.shape[i]
            
        self.tolerance = (max(self.max_coor) - min(self.min_coor))*1E-14

        # nicknames: xcoor, ycoor, xcoorv, ycoorv, etc
        for i in range(self.nsd):
            self.__dict__[self.dirnames[i]+'coor'] = self.coor[i]
            self.__dict__[self.dirnames[i]+'coorv'] = self.coorv[i]

        if self.nsd == 3:
            # make boundary coordinates for vectorization:
            xdummy, \
            self.ycoorv_xfixed_boundary, \
            self.zcoorv_xfixed_boundary = ndgrid(0, self.ycoor, self.zcoor)
            
            self.xcoorv_yfixed_boundary, \
            ydummy, \
            self.zcoorv_yfixed_boundary = ndgrid(self.xcoor, 0, self.zcoor)
            
            self.xcoorv_yfixed_boundary, \
            self.zcoorv_yfixed_boundary, \
            zdummy = ndgrid(self.xcoor, self.ycoor, 0)
            
    # could have _ in all variable names and define read-only
    # access via properties

    def string2griddata(s):
        """
        Turn a text specification of a grid into a dictionary
        with the grid data.
        For example,

        >>> s = "domain=[0,10] indices=[0:11]"
        >>> data = BoxGrid.string2griddata(s)
        >>> data
        {'dirnames': ('x', 'y'), 'division': [10], 'max': [10], 'min': [0]}

        >>> s = "domain=[0.2,0.5]x[0,2E+00] indices=[0:20]x[0:100]"
        >>> data = BoxGrid.string2griddata(s)
        >>> data
        {'dirnames': ('x', 'y', 'z'),
         'division': [19, 99],
         'max': [0.5, 2],
         'min': [0.2, 0]}

        >>> s = "[0,1]x[0,2]x[-1,1.5] [0:25]x[0:10]x[0:16]"
        >>> data = BoxGrid.string2griddata(s)
        >>> data
        {'dirnames': ('x', 'y', 'z'),
         'division': [24, 9, 15],
         'max': [1.0, 2.0, 1.5],
         'min': [0.0, 0.0, -1.0]}

        The data dictionary can be used as keyword arguments to the
        class UniformBoxGrid constructor.
        """
        
        domain  = r'\[([^,]*),([^\]]*)\]'
        indices = r'\[([^:,]*):([^\]]*)\]'
        import re
        d = re.findall(domain, s)
        i = re.findall(indices, s)
        nsd = len(d)
        if nsd != len(i):
            raise ValueError('could not parse "%s"' % s)
        kwargs = {}
        dirnames = ('x', 'y', 'z')
        for dir in range(nsd):
            if not isinstance(d[dir], (list,tuple)) or len(d[dir]) != 2 or \
               not isinstance(i[dir], (list,tuple)) or len(i[dir]) != 2:
                raise ValueError('syntax error in "%s"' % s)

            # old syntax (nx, xmin, xmax, ny, ymin, etc.):
            #kwargs[dirnames[dir]] = (float(d[dir][0]), float(d[dir][1]))
            #kwargs['n'+dirnames[dir]] = int(i[dir][1]) - int(i[dir][0]) # no of cells!
            kwargs['min'] = [float(d[dir][0]) for dir in range(nsd)]
            kwargs['max'] = [float(d[dir][1]) for dir in range(nsd)]
            kwargs['division'] = [int(i[dir][1]) - int(i[dir][0]) \
                                  for dir in range(nsd)]
            kwargs['dirnames'] = dirnames[:nsd]
        return kwargs
    string2griddata = staticmethod(string2griddata)

    def __getitem__(self, i):
        """
        Return access to coordinate array in direction no i, or direction
        name i, or return the coordinate of a point if i is an nsd-tuple.
        
        >>> g = UniformBoxGrid(x=(0,1), y=(-1,1), nx=2, ny=4)  # xy grid
        >>> g[0][0] == g.min[0]   # min coor in direction 0
        True
        >>> g['x'][0] == g.min[0]   # min coor in direction 'x'
        True
        >>> g[0,4]
        (0.0, 1.0)
        >>> g = UniformBoxGrid(min=(0,-1), max=(1,1), division=(2,4), dirnames=('y', 'z'))
        >>> g[1][0] == g.min[1]
        True
        >>> g['z'][0] == g.min[1]   # min coor in direction 'z'
        True
        """
        if isinstance(i, str):
            return self.coor[self.name2dirindex(i)]
        elif isinstance(i, int):
            if self.nsd > 1:
                return self.coor[i]     # coordinate array
            else:
                return self.coor[0][i]  # coordinate itself in 1D
        elif isinstance(i, (list,tuple)):
            return tuple([self.coor[k][i[k]] for k in range(len(i))])
        else:
            wrong_type(i, 'i', 'Must be str, int, tuple')
            

    def __setitem__(self, i, value):
        raise AttributeError('subscript assignment is not valid for '\
                             '%s instances' % self.__class__.__name__)

    def ncells(self, i):
        """Return no of cells in direction i."""
        # i has the meaning as in __getitem__. May be removed if not much used
        return len(self.coor[i])-1

    def name2dirindex(self, name):
        """
        Return direction index corresponding to direction name.
        In an xyz-grid, 'x' is 0, 'y' is 1, and 'z' is 2.
        In an yz-grid, 'x' is not defined, 'y' is 0, and 'z' is 1.
        """
        try:
            return self.dirnames.index(name)
        except ValueError:
            print name, 'is not defined'
            return None

    def dirindex2name(self, i):
        """Inverse of name2dirindex."""
        try:
            return self.dirnames[i]
        except IndexError:
            print i, 'is not a valid index'
            return None
    
    def ok(self):
        return True  # constructor init only => always ok

    def __len__(self):
        """Total number of grid points."""
        n = 1
        for dir in self.coor:
            n *= len(dir)
        return n

    def __repr__(self):
        s = self.__class__.__name__ + \
            '(min=%s, max=%s, division=%s, dirnames=%s)' % \
            (self.min_coor.tolist(),
             self.max_coor.tolist(),
             self.division, self.dirnames)
        return s

    def __str__(self):
        """Pretty print, using the syntax of init_fromstring."""
        domain = 'x'.join(['[%g,%g]' % (min_, max_) \
                           for min_, max_ in zip(self.min_coor, self.max_coor)])
        indices = 'x'.join(['[0:%d]' % div for div in self.division])
        return 'domain=%s  indices=%s' % (domain, indices)
                      
    def interpolator(self, point_values):
        """
        Given a self.nsd dimension array point_values with
        values at each grid point, this method returns a function
        for interpolating the scalar field defined by point_values
        at an arbitrary point.

        2D Example:
        given a filled array point_values[i,j], compute
        interpolator = grid.interpolator(point_values)
        v = interpolator(0.1243, 9.231)  # interpolate point_values
        
        >>> g=UniformBoxGrid(x=(0,2), nx=2, y=(-1,1), ny=2)
        >>> g
        UniformBoxGrid(x=(0,2), nx=2, y=(-1,1), ny=2)
        >>> def f(x,y): return 2+2*x-y

        >>> f=g.vectorized_eval(f)
        >>> f
        array([[ 3.,  2.,  1.],
               [ 5.,  4.,  3.],
               [ 7.,  6.,  5.]])
        >>> i=g.interpolator(f)
        >>> i(0.1,0.234)        # interpolate (not a grid point)
        1.9660000000000002
        >>> f(0.1,0.234)        # exact answer
        1.9660000000000002
        """
        args = self.coor
        args.append(point_values)
        # make use of wrap2callable, which applies ScientificPython
        return wrap2callable(args)

    def vectorized_eval(self, f):
        """
        Evaluate a function f (of the space directions) over a grid.
        f is supposed to be vectorized.

        >>> g = BoxGrid(x=(0,1), y=(0,1), nx=3, ny=3)
        >>> # f(x,y) = sin(x)*exp(x-y):
        >>> a = g.vectorized_eval(lambda x,y: sin(x)*exp(y-x))
        >>> print a
        [[ 0.          0.          0.          0.        ]
         [ 0.23444524  0.3271947   0.45663698  0.63728825]
         [ 0.31748164  0.44308133  0.6183698   0.86300458]
         [ 0.30955988  0.43202561  0.60294031  0.84147098]]

        >>> # f(x,y) = 2: (requires special consideration)
        >>> a = g.vectorized_eval(lambda x,y: zeros(g.shape)+2)
        >>> print a
        [[ 2.  2.  2.  2.]
         [ 2.  2.  2.  2.]
         [ 2.  2.  2.  2.]
         [ 2.  2.  2.  2.]]
        """
        a = f(*self.coorv)

        # check if f is really vectorized:
        try:
            msg = 'calling %s, which is supposed to be vectorized' % f.__name__
        except AttributeError:  # if __name__ is missing
            msg = 'calling a function, which is supposed to be vectorized'
        try:
            self.compatible(a)
        except (IndexError,TypeError), e:
            print 'e=',e, type(e), e.__class__.__name__
            raise e.__class__('BoxGrid.vectorized_eval(f):\n%s, BUT:\n%s' % \
                              (msg, e))
        return a
        
    def init_fromstring(s):
        data = UniformBoxGrid.string2griddata(s)
        return UniformBoxGrid(**data)
    init_fromstring = staticmethod(init_fromstring)

    def compatible(self, data_array, name_of_data_array=''):
        """
        Check that data_array is a NumPy array with dimensions
        compatible with the grid.
        """
        if not isinstance(data_array, ndarray):
            raise TypeError('data %s is %s, not NumPy array' % \
                            (name_of_data_array, type(data_array)))
        else:
            if data_array.shape != self.shape:
                raise IndexError("data %s of shape %s is not "\
                                 "compatible with the grid's shape %s" % \
                                 (name_of_data_array, data_array.shape,
                                  self.shape))
        return True # if we haven't raised any exceptions

    def iter(self, domain_part='all', vectorized_version=True):
        """
        Return iterator over grid points.
        domain_part = 'all':  all grid points
        domain_part = 'interior':  interior grid points
        domain_part = 'all_boundary':  all boundary points
        domain_part = 'interior_boundary':  interior boundary points
        domain_part = 'corners':  all corner points
        domain_part = 'all_edges':  all points along edges in 3D grids
        domain_part = 'interior_edges':  interior points along edges

        vectorized_version is true if the iterator returns slice
        objects for the index slice in each direction.
        vectorized_version is false if the iterator visits each point
        at a time (scalar version).
        """
        self.iterator_domain = domain_part
        self.vectorized_iter = vectorized_version
        return self

    def __iter__(self):
        # Idea: set up slices for the various self.iterator_domain
        # values. In scalar mode, make a loop over the slices and
        # yield the scalar value. In vectorized mode, return the
        # appropriate slices.
        
        self._slices = []  # elements meant to be slice objects
                    
        if self.iterator_domain == 'all':
            self._slices.append([])
            for i in range(self.nsd):
                self._slices[-1].append((i, slice(0, len(self.coor[i]), 1)))

        elif self.iterator_domain == 'interior':
            self._slices.append([])
            for i in range(self.nsd):
                self._slices[-1].append((i, slice(1, len(self.coor[i])-1, 1)))

        elif self.iterator_domain == 'all_boundary':
            for i in range(self.nsd):
                self._slices.append([])
                # boundary i fixed at 0:
                for j in range(self.nsd):
                    if j != i:
                        self._slices[-1].\
                           append((j, slice(0, len(self.coor[j]), 1)))
                    else:
                        self._slices[-1].append((i, slice(0, 1, 1)))
                # boundary i fixed at its max value:
                for j in range(self.nsd):
                    if j != i:
                        self._slices[-1].\
                           append((j, slice(0, len(self.coor[j]), 1)))
                    else:
                        n = len(self.coor[i])
                        self._slices[-1].append((i, slice(n-1, n, 1)))
                        
        elif self.iterator_domain == 'interior_boundary':
            for i in range(self.nsd):
                self._slices.append([])
                # boundary i fixed at 0:
                for j in range(self.nsd):
                    if j != i:
                        self._slices[-1].\
                           append((j, slice(1, len(self.coor[j])-1, 1)))
                    else:
                        self._slices[-1].append((i, slice(0, 1, 1)))
                # boundary i fixed at its max value:
                for j in range(self.nsd):
                    if j != i:
                        self._slices[-1].\
                           append((j, slice(1, len(self.coor[j])-1, 1)))
                    else:
                        n = len(self.coor[i])
                        self._slices[-1].append((i, slice(n-1, n, 1)))

        elif self.iterator_domain == 'corners':
            if self.nsd == 1:
                for i0 in (0, len(self.coor[0])-1):
                    self._slices.append([])
                    self._slices[-1].append((0, slice(i0, i0+1, 1)))
            elif self.nsd == 2:
                for i0 in (0, len(self.coor[0])-1):
                    for i1 in (0, len(self.coor[1])-1):
                        self._slices.append([])
                        self._slices[-1].append((0, slice(i0, i0+1, 1)))
                        self._slices[-1].append((0, slice(i1, i1+1, 1)))
            elif self.nsd == 3:
                for i0 in (0, len(self.coor[0])-1):
                    for i1 in (0, len(self.coor[1])-1):
                        for i2 in (0, len(self.coor[2])-1):
                            self._slices.append([])
                            self._slices[-1].append((0, slice(i0, i0+1, 1)))
                            self._slices[-1].append((0, slice(i1, i1+1, 1)))
                            self._slices[-1].append((0, slice(i2, i2+1, 1)))

        elif self.iterator_domain == 'all_edges':
            print 'iterator over "all_edges" is not implemented'
        elif self.iterator_domain == 'interior_edges':
            print 'iterator over "interior_edges" is not implemented'
        else:
            raise ValueError('iterator over "%s" is not impl.' % \
                             self.iterator_domain)
        
#    "def __next__(self):"
        """
        If vectorized mode:
        Return list of slice instances, where the i-th element in the
        list represents the slice for the index in the i-th space
        direction (0,...,nsd-1).

        If scalar mode:
        Return list of indices (in multi-D) or the index (in 1D).
        """
        if self.vectorized_iter:
            for s in self._slices:
                yield [slice_in_dir for dir, slice_in_dir in s]
        else:
            # scalar version
            for s in self._slices:
                slices = [slice_in_dir for dir, slice_in_dir in s]
                if len(slices) == 1:
                    for i in xrange(slices[0].start, slices[0].stop):
                        yield i
                elif len(slices) == 2:
                    for i in xrange(slices[0].start, slices[0].stop):
                        for j in xrange(slices[1].start, slices[1].stop):
                            yield [i, j]
                elif len(slices) == 3:
                    for i in xrange(slices[0].start, slices[0].stop):
                        for j in xrange(slices[1].start, slices[1].stop):
                            for k in xrange(slices[2].start, slices[2].stop):
                                yield [i, j, k]
                             

    def locate_cell(self, point):
        """
        Given a point (x, (x,y), (x,y,z)), locate the cell in which
        the point is located, and return
        1) the (i,j,k) vertex index
        of the "lower-left" grid point in this cell,
        2) the distances (dx, (dx,dy), or (dx,dy,dz)) from this point to
        the given point,
        3) a boolean list if point matches the
        coordinates of any grid lines. If a point matches
        the last grid point in a direction, the cell index is
        set to the max index such that the (i,j,k) index can be used
        directly for look up in an array of values. The corresponding
        element in the distance array is then set 0.
        4) the indices of the nearest grid point.
        
        The method only works for uniform grid spacing.
        Used for interpolation.

        >>> g1 = UniformBoxGrid(min=0, max=1, division=4)
        >>> cell_index, distance, match, nearest = g1.locate_cell(0.7)
        >>> print cell_index
        [2]
        >>> print distance
        [ 0.2]
        >>> print match
        [False]
        >>> print nearest
        [3]
        >>>
        >>> g1.locate_cell(0.5)
        ([2], array([ 0.]), [True], [2])
        >>>
        >>> g2 = UniformBoxGrid.init_fromstring('[-1,1]x[-1,2] [0:3]x[0:4]')
        >>> print g2.coor
        [array([-1.        , -0.33333333,  0.33333333,  1.        ]), array([-1.  , -0.25,  0.5 ,  1.25,  2.  ])]
        >>> g2.locate_cell((0.2,0.2))
        ([1, 1], array([ 0.53333333,  0.45      ]), [False, False], [2, 2])
        >>> g2.locate_cell((1,2))
        ([3, 4], array([ 0.,  0.]), [True, True], [3, 4])
        >>>
        >>>
        >>>
        """
        if isinstance(point, (int,float)):
            point = [point]
        nsd = len(point)
        if nsd != self.nsd:
            raise ValueError('point=%s has wrong dimension (this is a %dD grid!)' % \
                             (point, self.nsd))
        #index = zeros(nsd, int)
        index = [0]*nsd
        distance = zeros(nsd)
        grid_point = [False]*nsd
        nearest_point = [0]*nsd
        for i, coor in enumerate(point):
            # is point inside the domain?
            if coor < self.min_coor[i] or coor > self.max_coor[i]:
                raise ValueError(
                    'locate_cell: point=%s is outside the domain [%s,%s]' % \
                    point, self.min_coor[i], self.max_coor[i])
            index[i] = int((coor - self.min_coor[i])//self.delta[i])  # (need integer division)
            distance[i] = coor - (self.min_coor[i] + index[i]*self.delta[i])
            if distance[i] > self.delta[i]/2:
                nearest_point[i] = index[i] + 1
            else:
                nearest_point[i] = index[i]
            if abs(distance[i]) < self.tolerance:
                grid_point[i] = True
                nearest_point[i] = index[i]
            if (abs(distance[i] - self.delta[i])) < self.tolerance:
                # last cell, update index such that it coincides with the point
                grid_point[i] = True
                index[i] += 1
                nearest_point[i] = index[i]
                distance[i] = 0.0
                
        return index, distance, grid_point, nearest_point

    def interpolate(v0, v1, x0, x1, x):
        return v0 + (v1-v0)/float(x1-x0)*(x-x0)

    def gridline_slice(self, start_coor, direction=0, end_coor=None):
        """
        Compute start and end indices of a line through the grid,
        and return a tuple that can be used as slice for the
        grid points along the line.
        
        The line must be in x, y or z direction (direction=0,1 or 2).
        If end_coor=None, the line ends where the grid ends.
        start_coor holds the coordinates of the start of the line.
        If start_coor does not coincide with one of the grid points,
        the line is snapped onto the grid (i.e., the line coincides with
        a grid line).

        Return: tuple with indices and slice describing the grid point
        indices that make up the line, plus a boolean "snapped" which is
        True if the original line did not coincide with any grid line,
        meaning that the returned line was snapped onto to the grid.

        >>> g2 = UniformBoxGrid.init_fromstring('[-1,1]x[-1,2] [0:3]x[0:4]')
        >>> print g2.coor
        [array([-1.        , -0.33333333,  0.33333333,  1.        ]),
         array([-1.  , -0.25,  0.5 ,  1.25,  2.  ])]

        >>> g2.gridline_slice((-1, 0.5), 0)
        ((slice(0, 4, 1), 2), False)

        >>> g2.gridline_slice((-0.9, 0.4), 0)
        ((slice(0, 4, 1), 2), True)

        >>> g2.gridline_slice((-0.2, -1), 1)
        ((1, slice(0, 5, 1)), True)

        """
        
        start_cell, start_distance, start_match, start_nearest = \
                    self.locate_cell(start_coor)
        # If snapping the line onto to the grid is not desired, the
        # start_cell and start_match lists must be used for interpolation
        # (i.e., interpolation is needed in the directions i where
        # start_match[i] is False).
        
        start_snapped = start_nearest[:]
        if end_coor is None:
            end_snapped = start_snapped[:]
            end_snapped[direction] = self.division[direction] # highest legal index
        else:
            end_cell, end_distance, end_match, end_nearest = \
                      self.locate_cell(end_coor)
            end_snapped = end_nearest[:]
        # recall that upper index limit must be +1 in a slice:
        line_slice = start_snapped[:]
        line_slice[direction] = \
            slice(start_snapped[direction], end_snapped[direction]+1, 1)
        # note that if all start_match are true, then the plane
        # was not snapped
        return tuple(line_slice), not array(start_match).all()

        
    def gridplane_slice(self, value, constant_coor=0):
        """
        Compute a slice for a plane through the grid,
        defined by coor[constant_coor]=value.
        
        Return a tuple that can be used as slice, plus a boolean
        parameter "snapped" reflecting if the plane was snapped
        onto a grid plane (i.e., value did not correspond to
        an existing grid plane).
        """
        start_coor = self.min_coor.copy()
        start_coor[constant_coor] = value
        start_cell, start_distance, start_match, start_nearest = \
                    self.locate_cell(start_coor)
        start_snapped = [0]*self.nsd
        start_snapped[constant_coor] = start_nearest[constant_coor]
        # recall that upper index limit must be +1 in a slice:
        end_snapped = [self.division[i] for i in range(self.nsd)]
        end_snapped[constant_coor] = start_snapped[constant_coor]
        plane_slice = [slice(start_snapped[i], end_snapped[i]+1, 1) \
                       for i in range(self.nsd)]
        plane_slice[constant_coor] = start_nearest[constant_coor]
        return tuple(plane_slice), not start_match[constant_coor]
        

        
class BoxGrid(UniformBoxGrid):
    """
    Extension of class UniformBoxGrid to non-uniform box grids.
    The coordinate vectors (in each space direction) can have
    arbitrarily spaced coordinate values.

    The coor argument must be a list of nsd (number of
    space dimension) components, each component contains the
    grid coordinates in that space direction (stored as an array).
    """
    def __init__(self, coor, dirnames=('x', 'y', 'z')):
        
        UniformBoxGrid.__init__(self,
                                min=[a[0] for a in coor],
                                max=[a[-1] for a in coor],
                                division=[len(a)-1 for a in coor],
                                dirnames=dirnames)
        # override:
        self.coor = coor

    def __repr__(self):
        s = self.__class__.__name__ + '(coor=%s)' % self.coor
        return s

    def locate_cell(self, point):
        raise NotImplementedError('Cannot locate point in cells in non-uniform grids')
    
        
def _test(g, points=None):
    print 'g=%s' % str(g)
    # dump all the contents of a grid object:
    import scitools.misc;  scitools.misc.dump(g, hide_nonpublic=False)
    from numpy import zeros
    def fv(*args):
        # vectorized evaluation function
        return zeros(g.shape)+2
    def fs(*args):
        # scalar version
        return 2
    fv_arr = g.vectorized_eval(fv)
    fs_arr = zeros(g.shape)

    coor = [0.0]*g.nsd
    itparts = ['all', 'interior', 'all_boundary', 'interior_boundary',
               'corners']
    if g.nsd == 3:
        itparts += ['all_edges', 'interior_edges']
    for domain_part in itparts:
        print '\niterator over "%s"' % domain_part
        for i in g.iter(domain_part, vectorized_version=False):
            if isinstance(i, int):  i = [i]  # wrap index as list (if 1D)
            for k in range(g.nsd):
                coor[k] = g.coor[k][i[k]]
            print i, coor
            if domain_part == 'all':  # fs_arr shape corresponds to all points
                fs_arr[i] = fs(*coor)
        print 'vectorized iterator over "%s":' % domain_part
        for slices in g.iter(domain_part, vectorized_version=True):
            if domain_part == 'all':
                fs_arr[slices] = fv(*g.coor)
            # else: more complicated
            print slices
    # boundary slices...

    if points is not None:
        print '\n\nInterpolation in', g, '\n', g.coor
        for p in points:
            index, distance = g.locate_cell(p)
            print 'point %s is in cell %s, distance=%s' % (p, index, distance)
    

def _test2():
    g1 = UniformBoxGrid(min=0, max=1, division=4)
    _test(g1, [0.7, 0.5])
    spec = '[0,1]x[-1,2] with indices [0:3]x[0:2]'
    g2 = UniformBoxGrid.init_fromstring(spec)
    _test(g2, [(0.2,0.2), (1,2)])
    g3 = UniformBoxGrid(min=(0,0,-1), max=(1,1,1), division=(4,1,2))
    _test(g3)
    print 'g3=\n%s' % str(g3)
    print 'g3[Z]=', g3[Z]
    print 'g3[Z][1] =', g3[Z][1]
    print 'dx, dy, dz spacings:', g3.delta
    g4 = UniformBoxGrid(min=(0,-1), max=(1,1),
                        division=(4,2), dirnames=('y','z'))
    _test(g4)
    print 'g4["y"][-1]:', g4["y"][-1]
    
def _test4():
    from numpy import sin, zeros, exp
    # check vectorization evaluation:
    g = UniformBoxGrid(min=(0,0), max=(1,1), division=(3,3))
    try:
        g.vectorized_eval(lambda x,y: 2)
    except TypeError, msg:
        # fine, expect to arrive here
        print '*** Expected error in this test:', msg
    try:
        g.vectorized_eval(lambda x,y: zeros((2,2))+2)
    except IndexError, msg:
        # fine, expect to arrive here
        print '*** Expected error in this test:', msg

    a = g.vectorized_eval(lambda x,y: sin(x)*exp(y-x))
    print a
    a = g.vectorized_eval(lambda x,y: zeros(g.shape)+2)
    print a

        
if __name__ == '__main__':
    _test2()
    #_test4()



"""
This files is where all the options-related code is.
"""

# Standard library imports
import copy
import os
from optparse import OptionParser
import sys

# Local imports
from version import __version__

def silent_execute( string, return_stderr=True):
    """ Execute the given shell adding '> /dev/null' if under a posix OS 
        and '> nul' under windows.
    """
    if sys.platform.startswith('win') or return_stderr:
        return os.system(string + " > " + os.devnull)
    else:
        return os.system('%s >%s 2>%s' % (string, os.devnull,
                                                    os.devnull))


def verbose_execute(string):
    """ Execute getting errors """
    if os.system(string) != 0:
        raise RuntimeError('Unable to execute %r' % string)

# A dictionary describing the supported output type (as the keys of the
# dictionnary) and the figure type allowed for each.
allowed_types = {
        "html": ("png", "jpg") ,
        "rst" : ("png", "pdf", "ps", "jpg"), 
        "moin": ("png", "jpg") ,
        "trac" : ("png", "pdf", "ps", "jpg"), 
}

# Find out what output type we can provide (do we have latex ? epstopdf ?)
# FIXME: Have a look at mpl to see how they do this.
if not silent_execute("latex --help"):
    allowed_types.update({
        "tex" : ("pdf", "eps", "ps"),
        "dvi" : ("eps",),
        "ps"  : ("eps",),
        "eps" : ("eps",),
    })
    # Why the hell does epstopdf return 65280 !!
    if  silent_execute("epstopdf --help", return_stderr=False) in (0, 65280):
        allowed_types.update({
            "pdf" : ("pdf",),
            "tex" : ("pdf", "eps","ps"),
        })

if not silent_execute("pdflatex --help"):
    HAVE_PDFLATEX = True
else:
    HAVE_PDFLATEX = False

# Build a parser object
usage = """usage: %prog [options] pythonfile

Processes a python script and pretty prints the results using LateX. If 
the script uses "show()" commands (from pylab or scitools.easyviz) they
are caught by %prog and the resulting graphs are inserted in the output pdf.
Comments lines starting with "#!" are interprated as rst lines
and pretty printed accordingly in the pdf.
    By Gael Varoquaux"""

# Defaults are put to None and False in order to be able to track the changes.
option_parser = OptionParser(usage=usage, version="%prog " +__version__ )

option_parser.add_option("-o", "--outfile", dest="outfilename",
                help="write report to FILE", metavar="FILE")
option_parser.add_option("-x", "--noexecute",
                action="store_true", dest="noexecute", default=False,
                help="do not run the code, just extract the literate comments")
option_parser.add_option("-n", "--nocode",
                dest="nocode", action="store_true", default=False,
                help="do not display the source code")
option_parser.add_option("-d", "--double",
                dest="double", action="store_true", default=False,
                help="compile to two columns per page "
                     "(only for pdf or tex output)")
option_parser.add_option("-t", "--type", metavar="TYPE",
                action="store", type="string", dest="outtype",
                default=None,
                help="output to TYPE, TYPE can be " + 
                                        ", ".join(allowed_types.keys()))
option_parser.add_option("-f", "--figuretype", metavar="TYPE",
                action="store", type="string", dest="figuretype",
                default=None,
                help="output figure type TYPE  (TYPE can be of %s depending on report output type)" 
                % (", ".join(reduce(lambda x, y : set(x).union(y) , allowed_types.values()) )) )
option_parser.add_option("-c", "--commentchar",
                action="store", dest="commentchar", default="!",
                metavar="CHAR",
                help='literate comments start with "#CHAR" ')
option_parser.add_option("-l", "--latexliterals",
                action="store_true", dest="latexliterals",
                default=False,
                help='allow LaTeX literal comment lines starting with "#$" ')
option_parser.add_option("-e", "--latexescapes",
                action="store_true", dest="latexescapes",
                default=False,
                help='allow LaTeX math mode escape in code wih dollar signs ')
option_parser.add_option("-p", "--nopyreport",
                action="store_true", dest="nopyreport", default=False,
                help="disallow the use of #pyreport lines in the processed "
                     "file to specify options")
option_parser.add_option("-q", "--quiet",
                action="store_true", dest="quiet", default=False,
                help="don't print status messages to stderr")
option_parser.add_option("-v", "--verbose",
                action="store_true", dest="verbose", default=False,
                help="print all the message, including tex messages")
option_parser.add_option("-s", "--silent",
                dest="silent",action="store_true",
                default=False,
                help="""Suppress the display of warning and errors in the report""")
option_parser.add_option( "--noecho",
                dest="noecho",action="store_true",
                default=False,
                help="""Turns off the echoing of the output of the script on the standard out""")
option_parser.add_option("-a", "--arguments",
                action="store", dest="arguments",
                default=None, type="string", metavar="ARGS",
                help='pass the arguments "ARGS" to the script')

# Create default options
default_options, _not_used = option_parser.parse_args(args =[])
default_options._update_loose({
                                   'infilename': None,
                                   'outfile': None,
                               })

def parse_options(arguments, initial_options=copy.copy(default_options), 
                                    allowed_types=allowed_types):
    """ Parse options in the arguments list given.
        Return a dictionary containing all the options different specified,
        and only these, and the arguments.
        Returns outfilename without the extension ! (important)

    >>> parse_options(['-o','foo.ps','-s',])
    ({'outfilename': 'foo', 'outtype': 'ps', 'silent': True}, [])
    """
    (options, args) = option_parser.parse_args(args=arguments)
    if (options.outtype == None and 
            options.outfilename and 
            '.' in options.outfilename) :
        basename, extension = os.path.splitext(options.outfilename)
        if extension[1:] in allowed_types:
            options.outtype = extension[1:]
            options.outfilename = basename
    options_dict = options.__dict__
    initial_options_dict = initial_options.__dict__
    
    return diff_dict(options_dict, initial_options_dict), args

def diff_dict(dict1, dict2):
    """ Returns a dictionary with all the elements of dict1 that are not in
        dict 2.

    >>> diff_dict({1:2, 3:4}, {1:3, 3:4, 2:4})
    {1: 2}
    """
    return_dict = {}
    for key in dict1:
        if key in dict2:
            if not dict1[key] == dict2[key]:
                return_dict[key] = dict1[key]
        else:
            return_dict[key] = dict1[key]
    return return_dict






"""
This module provides a CodeHasher object that groups raw code lines in
full code blocks ready for execution.
"""

import token
import tokenize
import re
import StringIO
import platform

from options import parse_options

PYTHON_VERSION = int(''.join(str(s) for s in 
                            platform.python_version_tuple()[:2]))

def xreadlines(s):
    """ Helper function to use a string in the code hasher:
    blocks = iterblock(xreadlines('1\n2\n\n3'))
    """
    if  s and not s[-1]=="\n":
        s += "\n"
    return (line for line in StringIO.StringIO(s))


##############################################################################
class Token(object):
    """ A token object"""

    def __init__(self, token_desc):
        """ Builds a token object from the output of
            tokenize.generate_tokens"""
        self.type = token.tok_name[token_desc[0]]
        self.content = token_desc[1]
        self.start_row = token_desc[2][0]
        self.start_col = token_desc[2][1]
        self.end_row = token_desc[3][0]
        self.end_col = token_desc[3][1]

    def __repr__(self):
        return str((self.type, self.content))


##############################################################################
class CodeLine(object):
    """ An object representing a full logical line of code """
    string = ""
    open_symbols = {'{':0, '(':0, '[':0}
    closing_symbols = {'}':'{', ')':'(', ']':'['} 
    brakets_balanced = True
    end_col = 0
    last_token_type = ""
    complete = False
    options = {}

    def __init__(self, start_row):
        self.start_row = start_row
        self.end_row = start_row
    
    def append(self, token):
        """ Appends a token to the line while keeping the integrity of
            the line, and checking if the logical line is complete.
        """
        # The token content does not include whitespace, so we need to pad it
        # adequately
        token_started_new_line = False
        if token.start_row > self.end_row:
            self.end_col = 0
            token_started_new_line = True
        self.string += (token.start_col - self.end_col) * " " + token.content
        self.end_row = token.end_row
        self.end_col = token.end_col
        self.last_token_type = token.type

        # Keep count of the open and closed brakets.
        if token.type == 'OP':
            if token.content in self.open_symbols:
                self.open_symbols[token.content] += 1
            elif token.content in self.closing_symbols:
                self.open_symbols[self.closing_symbols[token.content]] += -1
            self.brakets_balanced = ( self.open_symbols.values() == [0, 0, 0] ) 
        
        self.complete = ( self.brakets_balanced 
                          and ( token.type in ('NEWLINE', 'ENDMARKER')
                                or ( token_started_new_line
                                      and token.type == 'COMMENT' )
                              )
                        )
        if ( token.type == 'COMMENT' 
                    and token_started_new_line 
                    and token.content[:10] == "#pyreport " ):
            self.options.update(parse_options(self.string[10:].split(" "))[0])

    def isnewblock(self):
        """ This functions checks if the code line start a new block.
        """
        # First get read of the leading empty lines:
        string = re.sub(r"\A([\t ]*\n)*", "", self.string)
        if re.match(r"elif|else|finally|except| |\t", string):
            return False
        else:
            return True        

    def __repr__(self):
        return('<CodeLine object, id %i, line %i, %s>'
                    % (id(self), self.start_row, repr(self.string) ) )


##############################################################################
class CodeBlock(object):
    """ Object that represents a full executable block """
    string = ""
    options = {}

    def __init__(self, start_row):
        self.start_row = start_row
        self.end_row = start_row

    def append(self, codeline):
        self.string += codeline.string
        self.options.update(codeline.options)

    def __repr__(self):
        return('<CodeBlock object, id %i, line %i, options %s\n%s>'
                    % (id(self), self.start_row, 
                            repr(self.options), self.string ) )


##############################################################################
class CodeHasher(object):
    """ Implements a object that transforms an iterator of raw code lines
        in an iterator of code blocks.

        Input:
            self.xreadlines: iterator to raw lines of code, such as 
                                 file.xreadlines()

        Output: Generators :
            self.itercodeblocks
            self.itercodelines
            self.itertokens
    """
    options = {}

    def __init__(self, xreadlines):
        """ The constructor takes as an argument an iterator on lines such 
            as the xreadline method of a file, or what is returned by the 
            xreadline function of this module.
        """
        self.xreadlines = xreadlines

    def next_line_generator(self):
        return self.xreadlines.next().expandtabs()

    def itercodeblocks(self):
        """ Returns a generator on the blocks of this code.
        """
        codeblock = CodeBlock(0)
        last_line_has_decorator = False
        for codeline in self.itercodelines():            
            if codeline.isnewblock() and not last_line_has_decorator :
                if codeblock.string:
                    self.options.update(codeblock.options)
                    codeblock.options.update(self.options)
                    yield codeblock
                codeblock = CodeBlock(codeline.start_row)
                codeblock.append(codeline)
                line_start = codeline.string.lstrip('\n')
                if line_start and line_start[0] == '@':
                        last_line_has_decorator = True
                        continue
# FIXME: I don't understand the purpose of this code. Until I don't have
# a test case that fail, I leave it commented out.
#                line_end = codeline.string.rstrip(" \n")
#                if line_end and line_end == ':' : 
#                    if codeblock.string:
#                        self.options.update(codeblock.options)
#                        codeblock.options.update(self.options)
#                        yield codeblock
#                    codeblock = CodeBlock(codeline.start_row)
            else:
                codeblock.append(codeline)
            last_line_has_decorator = False
        else:
            self.options.update(codeblock.options)
            codeblock.options.update(self.options)
            yield codeblock

    def itercodelines(self):
        """ Returns a generator on the logical lines of this code.
        """
        codeline = CodeLine(0)
        for token in self.itertokens():
            codeline.append(token)
            if codeline.complete:
                codeline.string = '\n'.join(s.rstrip(' ') 
                                    for s in codeline.string.split('\n'))
                yield codeline
                codeline = CodeLine(codeline.end_row + 1)
        if codeline.string:
            codeline.string = '\n'.join(s.rstrip(' ') 
                                    for s in codeline.string.split('\n'))
            yield codeline

    def itertokens(self):
        """ Returns a generator on the tokens of this code.
        """
        last_token = None
        for token_desc in tokenize.generate_tokens(self.next_line_generator):

            if PYTHON_VERSION < 26:
                yield Token(token_desc)
            else:
                # As of 2.6, tokenize.generate_tokens() chops newlines off
                # then end of comments and returns them as NL tokens. This
                # confuses the logic of the rest of pyreport, so we append
                # missing \n to COMMENT tokens, and gobble NL following a
                # comment.
                if token_desc[0] == tokenize.NL and \
                        last_token == tokenize.COMMENT:
                    last_token = token_desc[0]
                    continue
                else:
                    if token_desc[0] == tokenize.COMMENT \
                            and token_desc[1][-1] != '\n':
                        new_td = (token_desc[0], token_desc[1]+'\n', 
                                  token_desc[2], token_desc[3], token_desc[4])
                        token_desc = new_td

                    last_token = token_desc[0]
                    yield Token(token_desc)


iterblocks = lambda xreadlines: CodeHasher(xreadlines).itercodeblocks()






#!/usr/bin/env python

# Example taken from:
# http://www.mathworks.com/access/helpdesk/help/techdoc/ref/streamtube.html

from scitools.easyviz import *
from scipy import io

wind = io.loadmat('wind_matlab_v6.mat')
x = wind['x']
y = wind['y']
z = wind['z']
u = wind['u']
v = wind['v']
w = wind['w']

sx,sy,sz = ndgrid([80]*4,seq(20,50,10),seq(0,15,5),sparse=False)

setp(show=False)
streamtube(x,y,z,u,v,w,sx,sy,sz)
daspect([1,1,1])
view(3)
axis('tight')
shading('interp')
#camlight(); lighting('gouraud')
setp(show=True)
show()

figure()
# alternative syntax:
streamtube(x,y,z,u,v,w,sx,sy,sz,
           daspect=[1,1,1],
           view=3,
           axis='tight',
           shading='interp')

raw_input('Press Return key to quit: ')

#hardcopy('tmp_streamtube1.eps')
#hardcopy('tmp_streamtube1.png')



"""
This file is required only for loose coupling and avoiding import loops.
"""
# Author: Gael Varoquaux <gael dot varoquaux at normalesup dot org>
# Copyright: Gael Varoquaux
# License: BSD

__version__ = "0.3.4b"




#!/usr/bin/env python

# Example taken from:
# http://www.mathworks.com/access/helpdesk/help/techdoc/ref/streamline.html

from scitools.easyviz import *
from time import sleep
from scipy import io

wind = io.loadmat('wind_matlab_v6.mat')
x = wind['x']
y = wind['y']
z = wind['z']
u = wind['u']
v = wind['v']
w = wind['w']

setp(show=False)
sx,sy,sz = ndgrid([80]*4,seq(20,50,10),seq(0,15,5),sparse=False)
sl = streamline(x,y,z,u,v,w,sx,sy,sz)
sl.setp(linecolor='r')
view(3)
axis([60,140,10,60,-5,20])
setp(show=True)
show()


figure()
# alternative syntax:
sl = streamline(x,y,z,u,v,w,sx,sy,sz,
                linecolor='r',
                view=3,
                axis=[60,140,10,60,-5,20])


raw_input('Press Return key to quit: ')



from scitools.std import *
import time

xv, yv = ndgrid(linspace(-2,2,41),linspace(-1,1,41))
for n, t in enumerate(linspace(0, 4, 21)):
    values = exp(-0.2*t)*exp(-(2*(xv-t+1)**2 + 10*yv**2))
    surf(xv, yv, values, colormap=hot(24), colorbar='on',
         savefig='tmp_%03d.png' % n,
         zlim=[0,1], caxis=[0, 1])

# Note: this demo is very slow in matplotlib (3D)
raw_input()


